@@ -1,7 +1,7 @@
 /*
  *  This file is part of gpw.
  *
- *  Copyright (C) 2013-2016 Bruce Korb, all rights reserved.
+ *  Copyright (C) 2013-2018 Bruce Korb, all rights reserved.
  *  This is free software. It is licensed for use, modification and
  *  redistribution under the terms of the GNU General Public License,
  *  version 3 or later <http://gnu.org/licenses/gpl.html>
@@ -20,6 +20,12 @@
  *  with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
+#include <fcntl.h>
+#include <string.h>
+
+#include <sys/types.h>
+#include <sys/stat.h>
+
 /**
  * Hash a password id into an option search string.
  * The options associated with a password id are associated
@@ -32,7 +38,7 @@
 static char *
 make_pwid_mark(char const * name, size_t * len)
 {
-    char resbuf[256 / NBBY];
+    char resbuf[256 / NBBY]; // 256 bits (Number of Bits per BYte)
     char txtbuf[sizeof(long) + (sizeof(resbuf) * 3) / 2];
     struct sha256_ctx ctx;
     sha256_init_ctx(&ctx);
@@ -42,12 +48,118 @@ make_pwid_mark(char const * name, size_t * len)
     txtbuf[MARK_TEXT_LEN] = NUL;
 
     {
-        char * mark = scribble_get(id_mark_fmt_LEN + MARK_TEXT_LEN);
+        static unsigned long const mark_size =
+	    id_mark_fmt_LEN + MARK_TEXT_LEN + 10;
+        char * mark = scribble_get(mark_size);
         *len = sprintf(mark, id_mark_fmt, txtbuf);
         return mark;
     }
 }
 
+static char const *
+day_to_string(char const * day_str)
+{
+    static char time_buf[time_fmt_LEN + 4];
+
+    time_t day = strtoul(day_str, NULL, 10) * SECONDS_IN_DAY;
+    struct tm *tmday = localtime(&day);
+    if (tmday == NULL)
+        fserr(GNU_PW_MGR_EXIT_NO_MEM, "localtime", "");
+    strftime(time_buf, sizeof(time_buf), time_fmt, tmday);
+    return time_buf;
+}
+
+/**
+ * Find the next password id option in the config file that is not specified
+ * on the command line. It is a command line option if STATE_OPT(xx) is
+ * OPTST_DEFINED.
+ *
+ * @param scan      current position in config file data
+ * @param mark      the marker for "password id options"
+ * @param mark_len  the length of that marker
+ * @returns the character after the scan marker, or NULL
+ */
+static char const *
+next_pwid_opt(char const * scan, char const * mark, size_t mark_len)
+{
+    char * opt_text;
+
+    for (;;) {
+        scan = strstr(scan, mark);
+        if (scan == NULL)
+            return NULL;
+
+        scan += mark_len;
+        while (isspace((unsigned int)*scan))
+            scan++;
+        opt_text = strchr(scan, '>');
+        if (opt_text == NULL)
+            return NULL;
+        while (isspace((unsigned int)*++opt_text))  ;
+
+        /*
+         * If the found option type is in DEFINED state, then it was set
+         * on the command line and overrides whatever is in the config file.
+         */
+        switch (find_set_opt_cmd(opt_text)) {
+        case SET_CMD_LOGIN_ID:
+            if (STATE_OPT(LOGIN_ID) == OPTST_DEFINED)
+                continue;
+            break;
+
+        case SET_CMD_LENGTH:
+            if (STATE_OPT(LENGTH) == OPTST_DEFINED)
+                continue;
+            break;
+
+        case SET_CMD_CCLASS:
+            break; // always process this option
+
+        case SET_CMD_NO_PBKDF2:
+        case SET_CMD_USE_PBKDF2:
+            if (HAVE_OPT(REHASH))
+		continue;
+
+            if (strncmp(scan, date_z, date_z_LEN) == 0)
+                pbkdf2_date = day_to_string(scan + date_z_LEN);
+            else
+                pbkdf2_date = pw_undated;
+            break;
+
+        case SET_CMD_SPECIALS:
+            if (STATE_OPT(SPECIALS) == OPTST_DEFINED)
+                continue;
+            break;
+
+        case SET_CMD_SHARED:
+            if (STATE_OPT(SHARED) == OPTST_DEFINED)
+                continue;
+            break;
+
+        case SET_INVALID_CMD:
+        default:
+            goto no_next_pwid_opt;
+        }
+        return opt_text;
+    }
+
+    no_next_pwid_opt:
+    {
+        char * name = scribble_get(strlen(opt_text) + 1);
+        char * end  = name;
+        for (;;) {
+            unsigned char ch = (unsigned char)*(opt_text++);
+            if (! isalnum(ch))
+                break;
+            *(end++) = ch;
+        }
+        *end = NUL;
+        die(GNU_PW_MGR_EXIT_NO_CONFIG, bad_cfg_ent, mark, name);
+        /* NOTREACHED */
+    }
+    return NULL;
+}
+
 /**
  * set the options for a particular password id.
  * It modifies the \a optCookie field of \a DESC(CCLASS).
@@ -59,11 +171,17 @@ set_pwid_opts(char const * name)
 {
     char const * cfg_text = load_config_file();
 
-    /* Find the marker that separates the seeds from the
+    /*
+     * Find the marker that separates the seeds from the
      * password id options
      */
     char const * scan = strstr(cfg_text, pw_id_tag);
 
+    if (HAVE_OPT(REHASH)) {
+	pbkdf2_date = pw_today;
+	OPT_VALUE_PBKDF2 = OPT_VALUE_REHASH;
+    }
+
     if (scan != NULL) {
         size_t mark_len;
         char * mark = make_pwid_mark(name, &mark_len);
@@ -109,16 +227,33 @@ remove_opt(char const * txt, char const * mark, size_t m_len,
            set_opt_enum_t typ)
 {
     char * buf = (char *)(void *)txt;
+
     for (;;) {
-        char * p = strstr(buf, mark);
+        char * popt = strstr(buf, mark);
+        char * p    = popt;
+
         if (p == NULL)
             return;
 
-        if (find_set_opt_cmd(p + m_len) == typ) {
-            buf = p;
+        /*
+         * The marker may have more than just <pwtag id="..">, so
+         * scan over whatever else and past the closing '>'.
+         */
+        p = strchr(p, '>');
+        if (p == NULL)
+            return;
+
+        /*
+         * Convert the next part into the option type enumeration
+         * and see if it matches the one we're looking for.
+         * Spellings are "allowed" to vary, so it's not just a
+         * strncmp().
+         */
+        if (find_set_opt_cmd(++p) == typ) {
+            buf = popt;
             break;
         }
-        buf = p + m_len;
+        buf = p + id_mark_end_LEN;
     }
     {
         char * next = strstr(buf + m_len, pwtag_z);
@@ -155,8 +290,7 @@ removed_old_opts(char const * cfg_text, char const * name, char ** mark_p)
         res = true;
         remove_opt(cfg_text, mark, mark_len, SET_CMD_CCLASS);
     }
-
-    if (STATE_OPT(PBKDF2) == OPTST_DEFINED) {
+    if (HAVE_OPT(REHASH)) {
         res = true;
         remove_opt(cfg_text, mark, mark_len, SET_CMD_NO_PBKDF2);
         remove_opt(cfg_text, mark, mark_len, SET_CMD_USE_PBKDF2);
@@ -233,10 +367,12 @@ update_pwid_opts(char const * name)
             od->optArg.argString = save;
         }
 
-        if (STATE_OPT(PBKDF2) == OPTST_DEFINED) {
-            char const * how = ENABLED_OPT(PBKDF2) ? "use" : "no";
-            fprintf(fp, pwid_pbkdf2_fmt, mark, how,
-                    (unsigned int)OPT_VALUE_PBKDF2);
+	if (HAVE_OPT(REHASH)) {
+            unsigned int day = (unsigned int)
+                (time(NULL) / SECONDS_IN_DAY);
+
+            fprintf(fp, pwid_pbkdf2_fmt, mark, day,
+		    (uint32_t)OPT_VALUE_REHASH);
         }
 
         if (STATE_OPT(SPECIALS) == OPTST_DEFINED)
@@ -302,86 +438,8 @@ remove_pwid(char const * name)
 }
 
 /**
- * Find the next password id option in the config file that is not specified
- * on the command line.
- *
- * @param scan      current position in config file data
- * @param mark      the marker for "password id options"
- * @param mark_len  the length of that marker
- * @returns the character after the scan marker, or NULL
- */
-static char const *
-next_pwid_opt(char const * scan, char const * mark, size_t mark_len)
-{
-    for (;;) {
-        scan = strstr(scan, mark);
-        if (scan == NULL)
-            return NULL;
-
-        scan += mark_len;
-        while (isspace((unsigned int)*scan))  scan++;
-
-        /*
-         * If the found option type is in DEFINED state, then it was set
-         * on the command line and overrides whatever is in the config file.
-         */
-        switch (find_set_opt_cmd(scan)) {
-        case SET_CMD_LOGIN_ID:
-            if (STATE_OPT(LOGIN_ID) == OPTST_DEFINED)
-                continue;
-            break;
-
-        case SET_CMD_LENGTH:
-            if (STATE_OPT(LENGTH) == OPTST_DEFINED)
-                continue;
-            break;
-
-        case SET_CMD_CCLASS:
-            break; // always process this option
-
-        case SET_CMD_NO_PBKDF2:
-        case SET_CMD_USE_PBKDF2:
-            if (STATE_OPT(PBKDF2) == OPTST_DEFINED)
-                continue;
-            break;
-
-        case SET_CMD_SPECIALS:
-            if (STATE_OPT(SPECIALS) == OPTST_DEFINED)
-                continue;
-            break;
-
-        case SET_CMD_SHARED:
-            if (STATE_OPT(SHARED) == OPTST_DEFINED)
-                continue;
-            break;
-
-        case SET_INVALID_CMD:
-        default:
-            goto no_next_pwid_opt;
-        }
-        return scan;
-    }
-
-    no_next_pwid_opt:
-    {
-        char * name = scribble_get(strlen(scan) + 1);
-        char * end  = name;
-        for (;;) {
-            unsigned char ch = (unsigned char)*(scan++);
-            if (! isalnum(ch))
-                break;
-            *(end++) = ch;
-        }
-        *end = NUL;
-        die(GNU_PW_MGR_EXIT_NO_CONFIG, bad_cfg_ent, mark, name);
-        /* NOTREACHED */
-    }
-    return NULL;
-}
-
-/**
- * fix up the options.  We insert two options:  --load-opts and
- * --no-load-opts.  The former specifies the config file we decided upon,
+ * Insert two options:  --load-opts and --no-load-opts.
+ * The former specifies the config file we decided upon,
  * and the latter disables the processing of any other config files.
  * When the option processing is done, we'll choke and die if any other
  * config files got loaded.
@@ -390,7 +448,7 @@ next_pwid_opt(char const * scan, char const * mark, size_t mark_len)
  * @param av  pointer to argv argument to main
  */
 static void
-fix_options(int * ac, char *** av)
+insert_load_opts(int * ac, char *** av)
 {
     char *  fname = find_cfg_name();
     int     argc  = *ac + 3;
@@ -417,3 +475,63 @@ fix_options(int * ac, char *** av)
      */
     post_cfg_setting = OPT_VALUE_CCLASS;
 }
+
+/**
+ * fix up the options. If there is a "--config-file" option, we leave
+ * everything alone. Otherwise, we insert our own "--load-opts" and
+ * disable user's use of that option.
+ *
+ * @param ac  pointer to argc argument to main
+ * @param av  pointer to argv argument to main
+ */
+static void
+create_cfg_file(char * opt, char * opt1)
+{
+    struct stat sb;
+    int fno;
+
+    opt = strchr(opt, '=');
+    opt = (opt == NULL)
+        ? opt1
+        : opt + 1;
+
+    set_config_name(opt);
+    if (stat(opt, &sb) == 0)
+        return;
+
+    fno = open(opt, O_CREAT|O_WRONLY, S_IRWXU);
+    if (fno < 0)
+        fserr(GNU_PW_MGR_EXIT_INVALID, "open(O_CREAT)", opt);
+
+    if (fchmod(fno, S_IRUSR | S_IWUSR) != 0)
+        fserr(GNU_PW_MGR_EXIT_INVALID, "chmod", opt);
+
+    fno = close(fno);
+    if (fno != 0)
+        fserr(GNU_PW_MGR_EXIT_INVALID, "close", opt);
+}
+
+/**
+ * fix up the options. If there is a "--config-file" option, we leave
+ * everything alone. Otherwise, we insert our own "--load-opts" and
+ * disable user's use of that option.
+ *
+ * @param ac  pointer to argc argument to main
+ * @param av  pointer to argv argument to main
+ */
+static void
+fix_options(int * ac, char *** av)
+{
+    static char const conf[] = "--config";
+    int     argc  = *ac;
+    char ** argv  = *av;
+    while (--argc > 0) {
+        char * a = *++argv;
+        int    c = strncmp(a, conf, sizeof(conf) - 1);
+        if (c == 0) {
+            create_cfg_file(a, argv[1]);
+            return;
+        }
+    }
+    insert_load_opts(ac, av);
+}
